
#include "CodeBlock.h"
#include "HeapBlock.h"

namespace nanojit
{
	void CodeBlock::initialize(HeapBlock* terminator, NIns* end, CodeBlock* lower, CodeBlock* upper, CodeBlock* next) {

		NanoAssert(terminator != NULL);
		NanoAssert((((void*)upper == NULL) && ((void*)terminator == (void*)end)) || ((upper != NULL) && ((void*)upper == (void*)end)));

		_terminator = terminator; 
		_next = next;

		_end = end;
		
		_lower = lower;
		_upper = upper;

		_isFree = true;

		_codeStartOffsetInBytes = 0;
		_codeEndOffsetInBytes = 0;
	}

	bool CodeBlock::isBeforeInList(const CodeBlock* other) const {
		const CodeBlock* current = _next;
		while (current != NULL) {
			if (current == other)
				return true;

			current = current->_next;
		}

		return false;
	}

	uintptr_t CodeBlock::offsetFromCodeEndInBytes(const NIns* ins) const {
		NanoAssert(isInBlock(ins));
		return (uintptr_t(internalBufferEnd()) - _codeEndOffsetInBytes) - uintptr_t(ins);
	}

	uintptr_t CodeBlock::offsetFromCodeStartInBytes(const NIns* ins) const {
		NanoAssert(isInBlock(ins));
		return uintptr_t(ins) - uintptr_t(codeStart());
	}

	size_t CodeBlock::getMinimumFreeSizeForSplit(size_t minBlockSize) {
		return (minBlockSize * 2) + sizeOfBlockHeader() + alignof<CodeBlock>();
	}

	bool CodeBlock::canSplitBlock(size_t minBlockSize, CodeBlock* block, NIns* splitPoint) {
		// make sure the current block is big enough
		size_t minFreeSizeForSplit = getMinimumFreeSizeForSplit(minBlockSize);
		return (uintptr_t(splitPoint) - uintptr_t(block->internalBufferStart())) >= minFreeSizeForSplit;
	}

	CodeBlock* CodeBlock::splitBlock(size_t minBlockSize, CodeBlock* block, NIns* splitPoint) {
		NanoAssert(block->codeSize() > CodeBlock::sizeOfBlockHeader());
		NanoAssert(splitPoint > block->internalBufferStart());
		NanoAssert(canSplitBlock(minBlockSize, block, splitPoint));

		HeapBlock*	terminator	= block->_terminator;
		CodeBlock*	blockUpper	= block->_upper;
		CodeBlock*	blockLower	= block->_lower;
		NIns*		blockEnd	= block->internalBufferEnd();
		CodeBlock*	blockNext	= block->_next;

		size_t minSplitSize = minBlockSize + sizeof(CodeBlock);
		size_t codeBlockAlignMask = ~(alignof<CodeBlock>() - 1);
		uintptr_t newBlockAddress = (uintptr_t(splitPoint) - sizeof(CodeBlock)) & codeBlockAlignMask;
		if ((uintptr_t(block->internalBufferEnd()) - newBlockAddress) < minSplitSize) {
			newBlockAddress = (uintptr_t(block->internalBufferEnd()) - minSplitSize) & codeBlockAlignMask;
		}

		NanoAssert(newBlockAddress > uintptr_t(block->internalBufferStart()));
		NanoAssert(newBlockAddress < (uintptr_t(block->internalBufferEnd()) - sizeof(CodeBlock)));

		CodeBlock* newBlock = reinterpret_cast<CodeBlock*>(newBlockAddress);
		newBlock->initialize(terminator, blockEnd, block, blockUpper, NULL);
		NanoAssert(newBlock->internalBufferEnd() == block->internalBufferEnd());

		block->initialize(terminator, reinterpret_cast<NIns*>(newBlock), blockLower, newBlock, blockNext);
		NanoAssert(block->internalBufferEnd() == reinterpret_cast<NIns*>(newBlock));

		if (blockUpper != NULL)
			blockUpper->_lower = newBlock;

		NanoAssert((uintptr_t(block->internalBufferStart()) % alignof<NIns>()) == 0);
		NanoAssert((uintptr_t(block->internalBufferEnd()) % alignof<NIns>()) == 0);
		NanoAssert((uintptr_t(newBlock->internalBufferStart()) % alignof<NIns>()) == 0);
		NanoAssert((uintptr_t(newBlock->internalBufferEnd()) % alignof<NIns>()) == 0);

		return newBlock;
	}

	void CodeBlock::mergeCurrentWithUpper(CodeBlock* codeBlock) {
		NanoAssert(codeBlock->isFree());
		NanoAssert(codeBlock->upper()->isFree());
		NanoAssert(codeBlock->upper()->next() == NULL);

		CodeBlock* newLower = codeBlock->_lower;
		CodeBlock* newUpper = codeBlock->_upper->_upper;
		NIns* newEnd = codeBlock->_upper->_end;
		codeBlock->initialize(codeBlock->_terminator, newEnd, newLower, newUpper, codeBlock->_next);

		if (newUpper != NULL)
			newUpper->_lower = codeBlock;

		NanoAssert((codeBlock->lower() == NULL) || (codeBlock->lower()->upper() == codeBlock));
		NanoAssert((codeBlock->upper() == NULL) || (codeBlock->upper()->lower() == codeBlock));
	}
}

