
#include "nanojit.h"

namespace nanojit
{

#if defined(NANOJIT_VERBOSE)
	const char* LInsPrinter::accNames[] = {
		"v",    // (1 << 0) == ACCSET_VARS
		"t",    // (1 << 1) == ACCSET_TAGS
		"o",    // (1 << 2) == ACCSET_OTHER
		"?", "?", "?", "?", "?", "?", "?", "?",   //  3..10 (unused)
		"?", "?", "?", "?", "?", "?", "?", "?", "?", "?",   // 11..20 (unused)
		"?", "?", "?", "?", "?", "?", "?", "?", "?", "?",   // 21..30 (unused)
		"?"                                                 //     31 (unused)
	};

	char* LInsPrinter::formatAccSet(RefBuf* buf, AccSet accSet) {
		if (accSet == ACCSET_NONE) {
			VMPI_sprintf(buf->buf, ".none");
		} else if (accSet == ACCSET_ALL) {
			VMPI_sprintf(buf->buf, ".all");
		} else {
			char* b = buf->buf;
			b[0] = 0;
			// The AccSet may contain bits set for regions not used by the
			// embedding, if any have been specified via
			// (ACCSET_ALL & ~ACCSET_XYZ).  So only print those that are
			// relevant.
			for (int i = 0; i < EMB_NUM_USED_ACCS; i++) {
				if (accSet & (1 << i)) {
					VMPI_strcat(b, ".");
					VMPI_strcat(b, accNames[i]);
					accSet &= ~(1 << i);
				}
			}
			NanoAssert(VMPI_strlen(b) < buf->len);
		}
		return buf->buf;
	}

	char* LInsPrinter::formatImmI(RefBuf* buf, int32_t c) {
		if (-10000 < c && c < 10000) {
			VMPI_snprintf(buf->buf, buf->len, "%d", c);
		} else {
#if !defined NANOJIT_64BIT
			formatAddr(buf, (void*)c);
#else
			VMPI_snprintf(buf->buf, buf->len, "0x%x", (unsigned int)c);
#endif
		}
		return buf->buf;
	}

#if defined NANOJIT_64BIT
	char* LInsPrinter::formatImmQ(RefBuf* buf, uint64_t c) {
		if (-10000 < (int64_t)c && c < 10000) {
			VMPI_snprintf(buf->buf, buf->len, "%dLL", (int)c);
		} else {
			formatAddr(buf, (void*)c);
		}
		return buf->buf;
	}
#endif

	char* LInsPrinter::formatImmD(RefBuf* buf, double c) {
		VMPI_snprintf(buf->buf, buf->len, "%g", c);
		return buf->buf;
	}

	char* LInsPrinter::formatAddr(RefBuf* buf, void* p)
	{
		char*   name;
		int32_t offset;
		addrNameMap->lookupAddr(p, name, offset);

		if (name) {
			if (offset != 0) {
				VMPI_snprintf(buf->buf, buf->len, "%p %s+%d", p, name, offset);
			} else {
				VMPI_snprintf(buf->buf, buf->len, "%p %s", p, name);
			}
		} else {
			VMPI_snprintf(buf->buf, buf->len, "%p", p);
		}

		return buf->buf;
	}

	char* LInsPrinter::formatRef(RefBuf* buf, LIns *ref, bool showImmValue)
	{
		NanoAssert(ref != NULL);
		// Give 'ref' a name if it doesn't have one.
		const char* name = lirNameMap->lookupName(ref);
		if (!name) {
			name = lirNameMap->createName(ref);
		}

		// Put it in the buffer.  If it's an immediate, show the value if
		// showImmValue==true.  (This facility allows us to print immediate
		// values when they're used but not when they're def'd, ie. we don't
		// want "immi1/*1*/ = immi 1".)
		RefBuf buf2;
		if (ref->isImmI() && showImmValue) {
			VMPI_snprintf(buf->buf, buf->len, "%s/*%s*/", name, formatImmI(&buf2, ref->immI()));
		}
#ifdef NANOJIT_64BIT
		else if (ref->isImmQ() && showImmValue) {
			VMPI_snprintf(buf->buf, buf->len, "%s/*%s*/", name, formatImmQ(&buf2, ref->immQ()));
		}
#endif
		else if (ref->isImmD() && showImmValue) {
			VMPI_snprintf(buf->buf, buf->len, "%s/*%s*/", name, formatImmD(&buf2, ref->immD()));
		} 
		else if (ref->isImmF() && showImmValue) { // float is printed as a double
			VMPI_snprintf(buf->buf, buf->len, "%s/*%s*/", name, formatImmD(&buf2, ref->immF()));
		}
		else if (ref->isImmF4() && showImmValue) { // float4 is printed as an array of 4 double values
			RefBuf bufx, bufy, bufz, bufw;
			float4_t v = ref->immF4();
			VMPI_snprintf(buf->buf, buf->len, "%s/*%s,%s,%s,%s*/", name, formatImmD(&bufx,f4_x(v)),
				formatImmD(&bufy,f4_y(v)), formatImmD(&bufz,f4_z(v)), formatImmD(&bufw,f4_w(v)));
		}
		else {
			VMPI_snprintf(buf->buf, buf->len, "%s%s", ref->canOptimize() ? "" : "[*]", name);
		}

		return buf->buf;
	}

	char* LInsPrinter::formatIns(InsBuf* buf, LIns* i)
	{
		char *s = buf->buf;
		size_t n = buf->len;
		RefBuf b1, b2, b3, b4;
		LOpcode op = i->opcode();
		switch (op)
		{
		case LIR_immi:
			VMPI_snprintf(s, n, "%s = %s %s", formatRef(&b1, i, /*showImmValue*/false),
				getLirName(op), formatImmI(&b2, i->immI()));
			break;

#ifdef NANOJIT_64BIT
		case LIR_immq:
			VMPI_snprintf(s, n, "%s = %s %s", formatRef(&b1, i, /*showImmValue*/false),
				getLirName(op), formatImmQ(&b2, i->immQ()));
			break;
#endif

		case LIR_immf:
			VMPI_snprintf(s, n, "%s = %s %s", formatRef(&b1, i, /*showImmValue*/false),
				getLirName(op), formatImmD(&b2, i->immF()));
			break;

		case LIR_immf4:{
			RefBuf b5;
			float4_t v = i->immF4();
			VMPI_snprintf(s, n, "%s = %s %s,%s,%s,%s", formatRef(&b1, i, /*showImmValue*/false),
				getLirName(op), formatImmD(&b2, f4_x(v)),   formatImmD(&b3, f4_y(v)), formatImmD(&b4, f4_z(v)), formatImmD(&b5, f4_w(v)));
			break;
					   }

		case LIR_immd:
			VMPI_snprintf(s, n, "%s = %s %s", formatRef(&b1, i, /*showImmValue*/false),
				getLirName(op), formatImmD(&b2, i->immD()));
			break;

		case LIR_start:
			VMPI_snprintf(s, n, "%s", getLirName(op));
			break;

		case LIR_callv:
		case LIR_calli:
		CASE64(LIR_callq:)
		case LIR_calld:
		case LIR_callf:
		case LIR_callf4:
			{
				int32_t m = int32_t(n);     // Windows doesn't have 'ssize_t'
				m -= VMPI_snprintf(s, m, "%s = %s ( ", formatRef(&b1,   i), getLirName(op));

				for (uint32_t j = 0; j < i->callArgc(); j++) {
					s += VMPI_strlen(s);
					m -= VMPI_snprintf(s, m, "%s(%s) ", formatRef(&b2, i->callArgs()[j].value, false), regNames[REGNUM(i->callArgs()[j].reg)]);
					if (m < 0) break;
				}
				s += VMPI_strlen(s);
				m -= VMPI_snprintf(s, m, ")");
				break;
			}

		case LIR_parami:
		CASE64(LIR_paramq:)
		case LIR_paramf:
		case LIR_paramd:
		{
			Register reg = i->paramReg();
			VMPI_snprintf(s, n, "%s = %s %s%s", formatRef(&b1, i), getLirName(op), 
				reg == UnspecifiedReg ? "ar: " : "reg: ",
				reg == UnspecifiedReg ?
						formatImmI(&b2, i->getArGroup())
					:   regNames[REGNUM(reg)]);
			break;
		 }

		case LIR_spillfence:
		case LIR_regfence:
		case LIR_killfence: {

			VMPI_snprintf(
				s, n, 
				"%s = %s", 
				formatRef(&b1, i), 
				getLirName(i->opcode()));

			LIns**          values  = NULL;
			size_t          count   = 0;

			switch (op)
			{
				case LIR_spillfence:  values = i->spillFenceArray(); count = i->spillFenceCount(); break;
				case LIR_regfence:    values = i->registerFenceArray(); count = i->registerFenceCount(); break;
				case LIR_killfence:   values = i->killFenceArray(); count = i->killFenceCount(); break;
				default:			  NanoAssert(0); break;
			}

			b2.buf[0]   = '\0';
			for (unsigned int i = 0; i < count; i++) {
				LIns* paramValue = values[i];
				if (paramValue != NULL) {
					formatRef(&b2, paramValue, false);
					VMPI_snprintf(b3.buf, b3.len, " %s", b2.buf);
#if defined(WIN32)
					strcat_s(s, n, b3.buf);
#else
					strcat(s, b3.buf);
#endif
				}
			}
			break;
		}


		case LIR_label:
			VMPI_snprintf(s, n, "%s:", formatRef(&b1, i));
			break;

		case LIR_phireconcile:
		case LIR_phicapture: {

			VMPI_snprintf(
				s, n, 
				"%s = %s", 
				formatRef(&b1, i), 
				getLirName(i->opcode()));

			LIns**          values  = i->phiValues();
			size_t          count   = i->phiCount();

			b2.buf[0]   = '\0';
			for (unsigned int i = 0; i < count; i++) {
				LIns* paramValue = values[i];
				if (paramValue != NULL) {
					formatRef(&b2, paramValue, false);
					VMPI_snprintf(b3.buf, b3.len, " %s", b2.buf);
#if defined(WIN32)
					strcat_s(s, n, b3.buf);
#else
					strcat(s, b3.buf);
#endif
				}
			}
			break;
		}

		case LIR_jt:
		case LIR_jf:
			VMPI_snprintf(s, n, "%s %s -> %s", getLirName(op), formatRef(&b1, i->branchCondition()),
				i->branchTarget() ? formatRef(&b2, i->branchTarget()) : "unpatched");
			break;

		case LIR_j:
			NanoAssert(i->branchCondition() == NULL);
			VMPI_snprintf(s, n, "%s -> %s", getLirName(op),
				i->branchTarget() ? formatRef(&b1, i->branchTarget()) : "unpatched");
			break;

		case LIR_livei:
		case LIR_lived:
		case LIR_livef:
		case LIR_livef4:
		CASE64(LIR_liveq:)
			VMPI_snprintf(s, n, "%s %s", getLirName(op), formatRef(&b1, i->oprnd1()));
			break;

		case LIR_negi:
		case LIR_negd:
		case LIR_negf:
		case LIR_negf4:
		case LIR_absd:
		case LIR_absf:
		case LIR_absf4:
		case LIR_sqrtd:
		case LIR_sqrtf:
		case LIR_sqrtf4:
		case LIR_rsqrtf:
		case LIR_rsqrtf4:
		case LIR_recipf:
		case LIR_recipf4:
		case LIR_i2d:
		CASE64(LIR_q2d:)
		CASE64(LIR_d2q:)
		case LIR_ui2d:
		case LIR_i2f:
		case LIR_ui2f:
		case LIR_f4x:
		case LIR_f4y:
		case LIR_f4z:
		case LIR_f4w:
		case LIR_f2f4:
		case LIR_noti:
		CASE86(LIR_modi:)
		CASE64(LIR_i2q:)
		CASE64(LIR_ui2uq:)
		CASE64(LIR_q2i:)
		case LIR_d2i:
		case LIR_d2f:
		case LIR_f2d:
		case LIR_f2i:
		CASE64(LIR_dasq:)
		CASE64(LIR_qasd:)
			VMPI_snprintf(s, n, "%s = %s %s", formatRef(&b1, i), getLirName(op),
				formatRef(&b2, i->oprnd1()));
			break;

		case LIR_entry:
			VMPI_snprintf(s, n, "%s = %s", formatRef(&b1, i), getLirName(op));
			break;

		case LIR_exit: {
			VMPI_snprintf(s, n, "%s = %s", formatRef(&b1, i), getLirName(op));

			ParameterValue*     values  = i->exitParamData();
			unsigned int        count   = i->exitParamCount();
			b2.buf[0]   = '\0';
			for (unsigned int paramIndex = 0; paramIndex < count; paramIndex++) {
				LIns* paramValue = values[paramIndex].value;

				if (paramValue != NULL) {
					formatRef(&b2, paramValue, false);
					const char* regName = regNames[REGNUM(values[paramIndex].reg)];
					VMPI_snprintf(b3.buf, b3.len, " %s(%s)", b2.buf, regName);
#if defined(WIN32)
					strcat_s(s, n, b3.buf);
#else
					strcat(s, b3.buf);
#endif
				}
			}
			
			break;
		}

		case LIR_mullshii:
		case LIR_mulluhii:
		CASE64(LIR_mullshiq:)
		CASE64(LIR_mulluhiq:)
			VMPI_snprintf(s, n, "%s = %s %s",  formatRef(&b1, i), getLirName(op), formatRef(&b2, i->oprnd1()));
			break;

		case LIR_addi:  CASE64(LIR_addq:)
		case LIR_addci:CASE64(LIR_addcq:)
		case LIR_subi:  CASE64(LIR_subq:)
		case LIR_subbi: CASE64(LIR_subbq:)
		case LIR_muli: CASE64(LIR_mulq:)
		CASE86(LIR_diviu:)
		CASE86(LIR_divis:)
		case LIR_addd:
		case LIR_subd:
		case LIR_muld:
		case LIR_divd:
		case LIR_addf:
		case LIR_subf:
		case LIR_mulf:
		case LIR_divf:
		case LIR_addf4:
		case LIR_subf4:
		case LIR_mulf4:
		case LIR_divf4:
		case LIR_dotf4:
		case LIR_dotf3:
		case LIR_dotf2:
		case LIR_minf4:
		case LIR_maxf4:
		case LIR_cmpgtf4:
		case LIR_cmpgef4:
		case LIR_cmpltf4:
		case LIR_cmplef4:
		case LIR_cmpeqf4:
		case LIR_cmpnef4:
		case LIR_andi:       CASE64(LIR_andq:)
		case LIR_andci:      CASE64(LIR_andcq:)
		case LIR_ori:        CASE64(LIR_orq:)
		case LIR_xori:       CASE64(LIR_xorq:)
		case LIR_lshi:       CASE64(LIR_lshq:)
		case LIR_rshi:       CASE64(LIR_rshq:)
		case LIR_rshui:      CASE64(LIR_rshuq:)
		case LIR_rori:       CASE64(LIR_rorq:)
		case LIR_rcri:       CASE64(LIR_rcrq:)
		case LIR_eqi:        CASE64(LIR_eqq:)
		case LIR_nei:        CASE64(LIR_neq:)
		case LIR_lti:        CASE64(LIR_ltq:)
		case LIR_lei:        CASE64(LIR_leq:)
		case LIR_gti:        CASE64(LIR_gtq:)
		case LIR_gei:        CASE64(LIR_geq:)
		case LIR_ltui:       CASE64(LIR_ltuq:)
		case LIR_leui:       CASE64(LIR_leuq:)
		case LIR_gtui:       CASE64(LIR_gtuq:)
		case LIR_geui:       CASE64(LIR_geuq:)
		case LIR_eqd:
		case LIR_ltd:
		case LIR_led:
		case LIR_gtd:
		case LIR_ged:
		case LIR_eqf4:
		case LIR_eqf:
		case LIR_ltf:
		case LIR_lef:
		case LIR_gtf:
		case LIR_gef:
			VMPI_snprintf(s, n, "%s = %s %s, %s", formatRef(&b1, i), getLirName(op),
				formatRef(&b2, i->oprnd1()),
				formatRef(&b3, i->oprnd2()));
			break;

		CASE64(LIR_cmovq:)
		case LIR_cmovi:
		case LIR_cmovd:
		case LIR_cmovf:
		case LIR_cmovf4:
			VMPI_snprintf(s, n, "%s = %s %s ? %s : %s", formatRef(&b1, i), getLirName(op),
				formatRef(&b2, i->oprnd1()),
				formatRef(&b3, i->oprnd2()),
				formatRef(&b4, i->oprnd3()));
			break;

		case LIR_mullsi:
		case LIR_mullui:
		CASE64(LIR_mullsq:)
		CASE64(LIR_mulluq:)
			VMPI_snprintf(s, n, "%s = %s %s %s", formatRef(&b1, i), getLirName(op),
				formatRef(&b2, i->oprnd1()),
				formatRef(&b3, i->oprnd2()));
			break;

		case LIR_ldi:
		case LIR_ldri:
		CASE64(LIR_ldq:)
		CASE64(LIR_ldrq:)
		case LIR_ldd:
		case LIR_ldf:
		case LIR_ldf4:
		case LIR_lduc2ui:
		case LIR_ldus2ui:
		case LIR_ldrus2ui:
		case LIR_ldc2i:
		case LIR_lds2i:
		case LIR_ldrs2i:
		case LIR_ldf2d: {
			const char* qualStr;
			switch (i->loadQual()) {
			case LOAD_CONST:        qualStr = "/c"; break;
			case LOAD_NORMAL:       qualStr = "";   break;
			case LOAD_VOLATILE:     qualStr = "/v"; break;
			default: NanoAssert(0); qualStr = "/?"; break;
			}
			VMPI_snprintf(s, n, "%s = %s%s%s %s[%d]", formatRef(&b1, i), getLirName(op),
				formatAccSet(&b2, i->accSet()), qualStr, formatRef(&b3, i->oprnd1()),
				i->disp());
			break;
						}

		case LIR_sti:
		case LIR_stri:
		CASE64(LIR_stq:)
		CASE64(LIR_strq:)
		case LIR_std:
		case LIR_stf:
		case LIR_stf4:
		case LIR_sti2c:
		case LIR_sti2s:
		case LIR_stri2s:
		case LIR_std2f:
			VMPI_snprintf(s, n, "%s%s %s[%d] = %s", getLirName(op),
				formatAccSet(&b1, i->accSet()),
				formatRef(&b2, i->oprnd2()),
				i->disp(),
				formatRef(&b3, i->oprnd1()));
			break;

		case LIR_comment:
			VMPI_snprintf(s, n, "------------------------------ # %s", (char*)i->oprnd1());
			break;

		case LIR_swzf4:
			VMPI_snprintf(s, n, "%s %s, %d|%d|%d|%d", getLirName(LIR_swzf4),
				formatRef(&b1, i->oprnd1()),
				i->mask() >> 0 & 3,
				i->mask() >> 2 & 3,
				i->mask() >> 4 & 3,
				i->mask() >> 6 & 3);
			break;

		case LIR_copyi:
		CASE64(LIR_copyq:)
		case LIR_copyf:
		case LIR_copyd:
		case LIR_copyf4:
			VMPI_snprintf(s, n, "%s = %s %s", formatRef(&b1, i), getLirName(op), formatRef(&b2, i->oprnd1()));
			break;

		case LIR_setflgz:
		case LIR_setflgs:
		case LIR_setflgc:
		case LIR_setflgv:
			VMPI_snprintf(s, n, "%s = %s %s", formatRef(&b1, i), getLirName(op), formatRef(&b2, i->oprnd1()));
			break;

		case LIR_getflgz:
		case LIR_getflgs:
		case LIR_getflgc:
		case LIR_getflgv:
		case LIR_getflgnz:
		case LIR_getflgns:
		case LIR_getflgnc:
		case LIR_getflgnv:
			VMPI_snprintf(s, n, "%s = %s %s", formatRef(&b1, i), getLirName(op), formatRef(&b2, i->oprnd1()));
			break;

		case LIR_spillfenceall:
		case LIR_regfenceall:
		case LIR_nop:
			VMPI_snprintf(s, n, "%s", formatRef(&b1, i));
			break;

		case LIR_cvt2irf:
		case LIR_cvt2ird:
		case LIR_cvt2itf:
		case LIR_cvt2itd:
			VMPI_snprintf(s, n, "%s = %s %s", formatRef(&b1, i), getLirName(op), formatRef(&b2, i->oprnd1()));
			break;

		default:
			NanoAssertMsgf(0, "Can't handle opcode %s\n", getLirName(op));
			break;
		}
		return buf->buf;
	}
#endif // NANOJIT_VERBOSE
}

